import { Timer } from "./timer.js";
import { getVal, Var } from "./variables.js";

export class World {
  constructor(canvasId) {
    this.ctx = null;
    this.drawables = [];
    this.variables = [];
    this.backgroundColor = "white";
    this.timer = new Timer();
    this.targetFrameRate = 60;
    this.stepSize = 1000 / this.targetFrameRate;
    this.lastTick = this.timer.time();
    this.numTicks = 0;
    this.ctx = document.getElementById(canvasId).getContext("2d");
  }

  get width() {
    return this.ctx.canvas.width;
  }

  get height() {
    return this.ctx.canvas.height;
  }

  var(value) {
    let v = new Var(value);
    this.variables.push(v);
    return v;
  }

  drawable(thing) {
    this.drawables.push(thing);
    thing._world = this; // set world for access to renderer (TODO: revisit)
    return thing;
  }

  draw() {
    this.ctx.fillStyle = this.backgroundColor;
    this.ctx.beginPath();
    this.ctx.fillRect(0, 0, this.width, this.height);
    this.ctx.fill();
    for (let d of this.drawables.sort((a, b) => a._zIndex - b._zIndex)) {
      d.draw();
    }
  }

  tick() {
    this.numTicks++;
    for (let d of this.drawables) {
      d.update(this.numTicks);
    }
    for (let v of this.variables) {
      v.update(this.numTicks);
    }
  }

  get updatesPerSecond() {
    return (this.numTicks / this.timer.time()) * 1000;
  }

  loopStep() {
    window.requestAnimationFrame(() => this.loopStep());

    let curTime = this.timer.time();
    // tick appropriate number of times
    while (curTime - this.lastTick > this.stepSize) {
      this.lastTick += this.stepSize;
      this.tick();
    }
    this.draw();
  }

  // Canvas 2D /////////////////

  prepareDraw(drawable) {
    this._stroke = false;
    this._fill = false;
    let stroke = getVal(drawable._stroke);
    let fill = getVal(drawable._fill);
    if (stroke) {
      this.ctx.lineWidth = getVal(drawable._strokeWeight);
      this.ctx.strokeStyle = stroke.toStr ? stroke.toStr() : stroke;
      this._stroke = true;
    } else if (fill) {
      this.ctx.fillStyle = fill.toStr ? fill.toStr() : fill;
      this._fill = true;
    }
  }

  drawLine(from, to) {
    from = getVal(from);
    to = getVal(to);
    this.ctx.beginPath();
    this.ctx.moveTo(from.x, from.y);
    this.ctx.lineTo(to.x, to.y);
    this.ctx.stroke();
  }

  drawArc(center, radius, fromAngle, toAngle) {
    center = getVal(center);
    this.ctx.beginPath();
    this.ctx.arc(
      center.x,
      center.y,
      getVal(radius),
      getVal(fromAngle),
      getVal(toAngle),
    );
    if (this._stroke) this.ctx.stroke();
    if (this._fill) this.ctx.fill();
  }

  drawRect(center, width, height) {
    center = getVal(center);
    width = getVal(width);
    height = getVal(height);
    this.ctx.beginPath();
    this.ctx.rect(center.x - width / 2, center.y - height / 2, width, height);
    if (this._stroke) this.ctx.stroke();
    if (this._fill) this.ctx.fill();
  }

  drawPolygon(center, points) {
    center = getVal(center);
    this.ctx.beginPath();
    this.ctx.moveTo(center.x + points[0].x, center.y + points[0].y);
    for (let i = 1; i < points.length; i++) {
      this.ctx.lineTo(center.x + points[i].x, center.y + points[i].y);
    }
    this.ctx.lineTo(center.x + points[0].x, center.y + points[0].y);
    if (this._stroke) this.ctx.stroke();
    if (this._fill) this.ctx.fill();
  }
}
